/*                  Massachusetts Microelectronics Center           */
/*                                                                  */
/* Copyright Notice: All the modifications to the existing programs */
/*                   belong to Mass. Microelectronics Center.       */
/*                   Copyright and licensing requirements of the    */
/*                   original programs belong to the original       */
/*                   distributor.                                   */
/*                                                                  */
/* sccsid "@(#)N2.c	40.1      meg      6/29/88"               */

/*
**
**	Meg, the parent program to these routines interprets a grammar
**	for describing Mealy machine style finite state machines.  It is
**	used to generate a number of equivalent descriptions suitable for
**	automated CAD tools.  The tool these routines are aimed at is the
**	register transfer level simulation language N.2.  N2 has a C-like
**	syntax for control within each independent module.  It provides for
**	module local state and module interfaces (called ports).  A
**	collection of N2 modules can be interconnected to simulate a system.
**	
**	The routine PrintN2 is called by Meg after the input file has been
**	parsed into a data structure (described in ParserDefs.i) that is
**	input.  The file name of the input (and a flag indicating that the
**	input is really stdin) is also input.  PrintN2 will write to the
**	stdout a N2 process to implement the finite state machine Meg read.
**	It begins by replicating the data structure collected by the parser
**	so that it can reverse all lists without interfering with other meg
**	functions.  Lists need to be reversed beecause the parser stacks the
**	input structures.
**
**	The source file name is used in naming structures in the N2 output.
**	The effected structures are the state register internal to the module
**	called xxxxState (where the input file is xxxx.yyy such that yyy
**	contains no periods), the shadow state outputs Tracexxxx and the module
**	name xxxx.
**
**	There are a few introduced structure in addition to the three mentioned
**	above.  Two clock signals are input, SampleClock and AssertionClock.
**	The finite state machine examines its inputs on the trailing edge of
**	SampleClock and generates its outputs on the leading edge of
**	AssertionClock.  The first state found is assigned value zero and
**	this is the initial state.  There is also a procedure defined,
**	ClearAssertions, that resets all outputs zero time units before new
**	assertions are made.  This is done because N2 implicitly holds all
**	assignments until explicitly changed, but Meg input assumes all
**	values not explicitly assigned values each cycle are reset.
**
**	The N2 output is literred with N2 comments (same syntax as these
**	comments).  This is done because the user will 'grok' the input file
**	(which the user probably wrote and modifies) but must use the N2
**	code for N2 simulation and debugging.  The comments will have a
**	syntax closer to the input file than the N2 code because  N2 cannot
**	deal with 'don't care' bit values in a bit string value for a case
**	statement label - so nested ifs using bit srting masks are used.
**	The other reason that N2 looks different from the user's input is
**	that the user's input is run through the C preprocessor before Meg;
**	hence the user's defines (macros) are all expanded.  This will
**	typically effect input value definitions (Meg sees only 0, 1 and ?)
**	and sometimes sets of inputs or outputs driven together to indicate
**	a logical event.
**
**	Meg has a nasty habit of generating empty transition arc records at
**	the end of state cases with no explicit default transition.  Meg
**	later adjusts these to indicate sequential default arcs (ie. if no
**	default is given and no case is matched then goto the next state
**	listed in the input order).  This N2 representation does not encode
**	this - it assumes that all transitions are explicit.  This
**	discrepancy could lead to alot of trouble!!
**
*/

#include <stdio.h>
#include "ParserDefs.i"

/* #define DBG on */

#define LSS -1		/* used for string compare results */
#define EQL 0
#define GTR 1

#define TRUE 1		/* logic values */
#define FALSE 0


PrintN2( List, FileSource, FileName )
struct StateRecord *List;
int FileSource;
char *FileName;
{
    struct StateRecord *state, *s, *ReverseStateList();
    struct IdList *input, *bit;
    struct IdList *Inputs, *Outputs;
    struct IdList *VectorList, *v;
    struct CaseItem *arc;
    struct AssertList *value;
    char *Prefix, *Suffix, *t, *calloc();
    int NoStates, i, FieldWidth;

    if( FileSource == FALSE || FileName == (char *) 0 ) FileName = "n2.meg";
    for( t=(char *) 0, Suffix=Prefix=FileName; *Suffix != '\0'; Suffix++ )
        if( *Suffix == '.' ) t = Suffix;
    if( t != (char *) 0 ) { *t++ = '\0'; Suffix = t; };

/* sorry about the use of (type *) 0 but this is the coding convention
inherited from the originator of Meg
*/

    printf( "/* N2 test generation on input file %s.%s */\n\n",
            Prefix, Suffix );
    Inputs = (struct IdList *) 0;
    Outputs = (struct IdList *) 0;
    VectorList = (struct IdList *) 0;

/* create a copy of the parser data structure with reversed lists
*/
    state = ReverseStateList( List, &Inputs, &Outputs );

    for( printf( "\n/* States:" ), s = state;
         s != (struct StateRecord *) 0; s = s->next )
         printf( " %s", s->name );
    printf( " */\n\n" );

/* assign values to the states and introduce the concat macro (used to reduce
   and simplify bit string construction strings
*/
    for( printf( "macro\n" ), NoStates = 0, s = state;
         s != (struct StateRecord *) 0; s = s->next )
         printf( "    %s := %d&,\n", s->name, NoStates++ ); 
    printf( "    _ := concat &;\n" );
    /* compute the state register width in bits */
    for( i = NoStates, FieldWidth = 1; i > 1; i = i/2, FieldWidth++ );

    for( printf( "\n/* Inputs:" ), bit = Inputs;
                     bit != (struct IdList *) 0; bit = bit->next )
         printf( " %s", bit->name );
    printf( " */\n\n" );

/* Define the input only ports and the clocks
*/
    for( printf( "port\n" ), bit = Inputs;
                     bit != (struct IdList *) 0; bit = bit->next )
         printf( "    %s'input,\n", bit->name );
    printf( "    SampleClock'input,\n    AssertionClock'input;\n" );

    for( printf( "\n/* Outputs:" ), bit = Outputs;
                     bit != (struct IdList *) 0; bit = bit->next )
         printf( " %s", bit->name );
    printf( " */\n\n" );

/* Define the output only ports including the state tracing lines
*/
    for( printf( "port\n" ), bit = Outputs;
                     bit != (struct IdList *) 0; bit = bit->next )
         printf( "    %s'output,\n", bit->name );
    printf( "    Trace%s<%d> (0)'output;\n", Prefix, FieldWidth );

/* Define the state register
*/
    printf( "\nstate\n    %sState<%d> (0);\n", Prefix, FieldWidth );

/* This procedure resets all outputs to the assumed unasserted value, '0'.
   It is called before any assertions are done during an transition.
   Because no simulation time elapses and because a next forces assignment
   ordering, if some values are incorrectly cleared their change will not
   be seen as a glitch.
*/
    for( printf( "\nprocedure ClearAssertions := (\n" ), bit = Outputs;
                     bit != (struct IdList *) 0; bit = bit->next )
         printf( "    %s = 0;\n", bit->name );
    printf( "    next;\n)\n" );

/* introduce the finite state module name and clock trigger and case on
the state register: the body is nested if's within state
*/
    printf( "\nwhen %s ( SampleClock:trail ) := (\n", Prefix );
    printf( "\n    case %sState\n", Prefix );
   
    for( ; state != (struct StateRecord *) 0; state = state->next )
    {
/* start comments for more Meg like description of operation */
      printf( "\n/* STATE */ %s :\t\t/* GRAPH NODE N:%s\n\n\tcase on:",
                  state->name, state->name );
      for( input = state->selector;
           input != (struct IdList *) 0;
           input = input->next )
        printf( " %s", input->name );
      for( i=0, arc = state->cases;
           arc != (struct CaseItem *) 0;
           i++, arc = arc->next )
      {
        if( arc->bits == (struct IdList *) 0 && arc->StateName == (char *) 0 )
            continue;
        printf( "\n\t\t" );
        if( arc->bits == (struct IdList *) 0 )
	    printf( "default" );
        else
            for( bit = arc->bits;
                 bit != (struct IdList *) 0;
                 bit = bit->next )
                printf( "%s ", bit->name );
        printf( "=> %s\tGRAPH ARC A:%s_%d\n\t\t\tOutputs:",
		arc->StateName, state->name, i );
        for( value = arc->Assert;
             value != (struct AssertList *) 0;
             value = value->next )
          printf( " %s=%c", value->name, value->bit );
      }
      printf( "\n*/\n\t" );
/* end comment */

/* This manipulation results from the 'don't cares' in Meg and the lack of
same in N2.  The bit vector case'd on in this state is constructed in a
separate string so it can be reprinted easily in each transition definition.
A bit vector is joined using the '_' macro for the N2 'concat' operator.
*/
      for( i = 0, input = state->selector;
           input != (struct IdList *) 0;
           i += 3, input = input->next )
          i += StringLength( input->name );
      v = (struct IdList *) calloc( 1, sizeof(struct IdList));
      v->next = VectorList;
      v->name = (char *) calloc( i+1, sizeof(char) );
      VectorList = v;
      for( i = 0, input = state->selector;
           input != (struct IdList *) 0;
           i += 3, input = input->next )
      {
          StringCopyFromTo( input->name, &(v->name[i]) );
          i += StringLength( input->name );
          StringCopyFromTo( " _ ", &(v->name[i]) );
      }
      v->name[i-3] = '\0';

/* Each transition arc is tested for by bitwise and'ing the vector of input
signals against a mask and comparing to the arc value (post mask).
The mask has a 0 for each don't care in the Meg specification and a 1 for
each non-don't care.  The arc value has each don't care replaced with a 0.
Again the '_' macro joins bits into a vector.
*/
      for( arc = state->cases;
           arc != (struct CaseItem *) 0;
           arc = arc->next )
      {
/* If the transition has no values and no next state ignore it!
*/
       if( arc->bits == (struct IdList *) 0 && arc->StateName == (char *) 0 )
           continue;
/* If the transition has no values but does have a next state it is assumed
to be the trailing default arc and will be the statement closing the else
ie. no 'if' prefix is generated
*/
       if( arc->bits != (struct IdList *) 0 )
       {
           printf( "if %s\n\t\tand ", v->name );
           for( bit = arc->bits;
                bit != (struct IdList *) 0;
                bit = bit->next )
               printf( "%s%s", (CompStr(bit->name,"-")==EQL)? "0" : "1",
                       (bit->next==(struct IdList *) 0)? " " : " _ " );
           printf( "eql " );
           for( bit = arc->bits;
                bit != (struct IdList *) 0;
                bit = bit->next )
               printf( "%s%s", (CompStr(bit->name,"-")==EQL)? "0" : bit->name,
                       (bit->next==(struct IdList *) 0)? " " : " _ " );
        }
        printf( " (\n" );

/* The body of an if represents the arc action.  First all outputs are
reset, then after zero time the new assertions are made.  This models the
assumed finite state machine implementation.  The resets and assertions are
delayed until the leading edge of the assertion clock to model computation
delay.  The new state value is echoed on the state tracing lines.
*/
        printf( "\t\twait( AssertionClock:lead );\n" );
        printf( "\t\tClearAssertions;\n" );
        printf( "\t\t%sState = %s;\n", Prefix, arc->StateName );
        printf( "\t\tTrace%s = %s;\n", Prefix, arc->StateName );
        for( value = arc->Assert;
             value != (struct AssertList *) 0;
             value = value->next )
          printf( "\t\t%s = %c;\n", value->name, value->bit );

/* an else cause is included if their is another explicit arc in the list:
Meg unfortunately generates an empty transition at the end of the list to
represent an implicit default transition to the next sequential state.
This N2 representation does no such thing.  All transitions must be
explicit.
*/
        printf( "\t)\n%s",
                (arc->next != (struct CaseItem *) 0
                 && arc->next->StateName != (char *) 0
                ) ? "\telse " : "" );
      }
      printf( "\n" );
    }
    printf( "    esac\n)\n" );
}
/* end PrintN2 code */


/* the remaining code implements tools for the N2 coding
the list reversing code is broken down by the types of lists
(see ParserDefs.i) and some string manipulating routines are given
*/

struct StateRecord *
ReverseStateList( statelist, inidp, outidp )
struct StateRecord *statelist;
struct IdList **inidp, **outidp;
/*
Reverse the list so that it is in the order specified by the source.
The first parameter is the list to be reversed, the other two are pointers
to list pointers.  These last two are used to construct symbol tables for
the input only and output only signals.
*/
{
    struct StateRecord * src, * tgt, * temp;
    struct IdList * ReverseIdList(), * id1, * t, ** id2p;
    struct CaseItem * ReverseCases(), * cs;
    struct AssertList * as;
    char * calloc();
    int r;

    tgt = (struct StateRecord *) 0;
    for( src = statelist; src != (struct StateRecord *) 0; src = src->next )
    {
	temp = tgt;
        tgt = (struct StateRecord *) calloc( 1, sizeof(struct
                                             StateRecord));
        tgt->selector = ReverseIdList( src->selector );
	tgt->name = src->name;
	tgt->cases = ReverseCases( src->cases );
	tgt->next = temp;
    }
/* the rest of this constructs a symbol table of the input only and output
only signals - the list is ordered
*/
#ifdef DBG
    printf( "<start I/O list collect>\n" );
#endif
    for( src = tgt; src != (struct StateRecord *) 0; src = src->next)
    {
#ifdef DBG
        printf( "<S:%s>\n", src->name );
#endif
        for( id1 = src->selector; id1 != (struct IdList *) 0; id1 = id1->next)
        {
            for( id2p = inidp; (*id2p) != (struct IdList *) 0; 
                                               id2p = &((*id2p)->next) )
            {
                r = CompStr( id1->name, (*id2p)->name );
                if( r == EQL ) break;
                if( r == LSS )
                {
                    t = (struct IdList *) calloc( 1, sizeof(struct IdList));
                    t->name = id1->name;
                    t->next = (*id2p);
                    *id2p = t;
                    break;
                }
            }
            if( (*id2p) == (struct IdList *) 0 )
            {
                (*id2p) = (struct IdList *) calloc( 1, sizeof(struct IdList));
                (*id2p)->name = id1->name;
                (*id2p)->next = (struct IdList *) 0;
            }
        }

        for( cs = src->cases; cs != (struct CaseItem *) 0; cs = cs->next)
        {
            for( as = cs->Assert; as != (struct AssertList *) 0; as = as->next )
            {
                for( id2p = outidp; (*id2p) != (struct IdList *) 0; 
                                                   id2p = &((*id2p)->next) )
                {
                    r = CompStr( as->name, (*id2p)->name );
                    if( r == EQL ) break;
                    if( r == LSS )
                    {
                        t = (struct IdList *) calloc( 1, sizeof(struct IdList));
                        t->name = as->name;
                        t->next = (*id2p);
                        *id2p = t;
                        break;
                    }
                }
                if( (*id2p) == (struct IdList *) 0 )
                {
                    (*id2p) = (struct IdList *)
                                      calloc( 1, sizeof(struct IdList));
                    (*id2p)->name = as->name;
                    (*id2p)->next = (struct IdList *) 0;
                }
            }
        }
    }
    return( tgt );
}

struct IdList *
ReverseIdList( idlist )
struct IdList * idlist;
/* reverse the list so that it is in the order found in the source */
{
    struct IdList * src, * tgt, * temp;
    char * calloc();

    tgt = (struct IdList *) 0;
    for( src = idlist; src != (struct IdList *) 0; src = src->next )
    {
        temp = tgt;
        tgt = (struct IdList *) calloc( 1, sizeof(struct IdList));
        tgt->name = src->name;
        tgt->next = temp;
    }
    return( tgt );
}


struct CaseItem *
ReverseCases( caselist )
struct CaseItem * caselist;
/* reverse the list so that it is in the order found in the source */
{
    struct CaseItem * src, * tgt, * temp;
    struct AssertList * ReverseValues();
    struct IdList * ReverseIdList();
    char * calloc();

    tgt = (struct CaseItem *) 0;
    for( src = caselist; src != (struct CaseItem *) 0; src = src->next )
    {
        temp = tgt;
        tgt = (struct CaseItem *) calloc( 1, sizeof(struct CaseItem));
        tgt->bits = ReverseIdList( src->bits );
        tgt->Assert = ReverseValues( src->Assert );
        tgt->StateName = src->StateName;
        tgt->next = temp;
    }
    return( tgt );
}

struct AssertList *
ReverseValues( valuelist )
struct AssertList * valuelist;
/* reverse the list so that it is in the order found in the source */
{
    struct AssertList * src, * tgt, * temp;
    char * calloc();

    tgt = (struct AssertList *) 0;
    for( src = valuelist; src != (struct AssertList *) 0; src = src->next )
    {
        temp = tgt;
        tgt = (struct AssertList *) calloc( 1, sizeof(struct AssertList));
        tgt->bit = src->bit;
        tgt->name = src->name;
        tgt->next = temp;
    }
    return( tgt );
}

int
CompStr( s1, s2 )
char * s1, *s2;
{
    while( *s1 != '\0' && *s1 == *s2 ) { s1++; s2++; };
    if( *s1 == *s2 ) return( EQL );
    else if( *s1 < *s2 ) return( LSS );
         else return( GTR );
}

int
StringLength( s )
char * s;
{
/* returns no. of characters in string pointed to by s not counting
   the trailing \0 character - returns 0 if s is null */

    int l;

    if( s == (char *) 0 ) return( 0 );
    for( l = 0; *s != '\0'; l++, s++ );
    return( l );
}

StringCopyFromTo( s1, s2 )
char * s1, * s2;
{
/* copies the characters in the string pointed to be s1 into the string
   pointed to by s2 including the \0 at the end
   NOTE: s2 MUST BE AT LEAST AS LONG AS s1 */

    if( s1 == (char *) 0 || s2 == (char *) 0 )
    {
	printf( "DEATH in N2Print\n" );
        exit(1);
    }
    for( ; *s1 != '\0'; *s2++ = *s1++ );
    *s2 = *s1;
}
